function [h,pValue,stat,cValue,mles] = jcontest(Y,r,test,Cons,varargin)
%JCONTEST Johansen constraint test
%
% Syntax:
%
%   [h,pValue,stat,cValue,mles] = jcontest(Y,r,test,Cons)
%   [h,pValue,stat,cValue,mles] = jcontest(Y,r,test,Cons,param,val,...)
%
% Description:
%
%   JCONTEST tests linear constraints on either the error-correction speeds
%   A or the cointegration space spanned by B in the reduced-rank VEC(q)
%   model of y(t):
%
%       (1-L)y(t) = A*B'*y(t-1)
%                   + B1*(1-L)y(t-1) + ... + Bq*(1-L)y(t-q)
%                   + D*x
%                   + e(t)
%
%   where L is the lag operator Ly(t) = y(t-1). Null hypotheses specifying
%   constraints on A or B are tested against the alternative H(r) of
%   cointegration rank less than or equal to r, without the constraints.
%   The tests also produce maximum likelihood estimates of the parameters
%   in the VEC(q) model, subject to the constraints.
%
% Input Arguments:
%
%   Y - numObs-by-numDims matrix representing numObs observations of a
%       numDims-dimensional time series y(t), with the last observation the
%       most recent. Observations containing NaN values are removed.
%       Initial values for lagged variables in VEC model estimation are
%       taken from the beginning of the data.
%
%   r - Scalar or vector of integers between 1 and numDims-1, inclusive,
%       specifying the common rank of A and B, as inferred by JCITEST.
%
%   test - String or cell vector of strings specifying the type of tests to
%       be performed. Values are:
%
%       'ACon'  Test linear constraints on A.
%
%       'AVec'  Test specific vectors in A.
%
%       'BCon'  Test linear constraints on B.
%
%       'BVec'  Test specific vectors in B.
%
%   Cons - Matrix or cell vector of matrices specifying test constraints.
%       For constraints on B, the number of rows in each matrix, numDims1,
%       is the number of dimensions in the data, numDims, unless 'model' is
%       'H*' or 'H1*', in which case numDims1 = numDims+1 and constraints
%       include the restricted deterministic term in the model.
%
%       TEST    CONS
%
%       'ACon'  numDims-by-numCons matrix R specifying numCons constraints
%               on A given by R'*A = 0. numCons must not exceed numDims-r.
%
%       'AVec'  numDims-by-numCons matrix specifying numCons of the error-
%               correction speed vectors in A. numCons must not exceed r.
%
%       'BCon'  numDims1-by-numCons matrix R specifying numCons constraints
%               on B given by R'*B = 0. numCons must not exceed numDims-r.
%
%       'BVec'  numDims1-by-numCons matrix specifying numCons of the
%               cointegrating vectors in B. numCons must not exceed r.
%
% Optional Input Parameter Name/Value Pairs:
%
%   NAME        VALUE
%
%   'model' 	String or cell vector of strings specifying the form of the
%               deterministic components of the VEC(q) model of y(t).
%               Values of 'model' are those considered by Johansen [3]:
%
%               VALUE       FORM OF A*B'*y(t-1) + D*x
%
%               o 'H2'      A*B'*y(t-1). There are no intercepts or trends
%                           in the cointegrating relations and there are no
%                           trends in the data. This model is only
%                           appropriate if all series have zero mean.
%
%               o 'H1*'     A*(B'*y(t-1) + c0). There are intercepts in the
%                         	cointegrating relations and there are no
%                           trends in the data. This model is appropriate
%                           for nontrending data with nonzero mean.
%
%               o 'H1'      A*(B'*y(t-1) + c0) + c1. There are intercepts
%                           in the cointegrating relations and there are
%                           linear trends in the data. This is a model of
%                           "deterministic cointegration," where the
%                           cointegrating relations eliminate both
%                           stochastic and deterministic trends in the
%                           data. This is the default value.
%
%               o 'H*'      A*(B'*y(t-1) + c0 + d0*t) + c1. There are
%                           intercepts and linear trends in the
%                           cointegrating relations and there are linear
%                           trends in the data. This is a model of
%                           "stochastic cointegration," where the
%                           cointegrating relations eliminate stochastic
%                           but not deterministic trends in the data.
%
%               o 'H'       A*(B'*y(t-1) + c0 + d0*t) + c1 + d1*t. There
%                           are intercepts and linear trends in the
%                           cointegrating relations and there are quadratic
%                           trends in the data. Unless quadratic trends are
%                           actually present in the data, this model may
%                           produce good in-sample fits but poor out-of-
%                           sample forecasts.
%
%               Deterministic terms outside of the cointegrating relations,
%               c1 and d1, are identified by projecting constant and linear
%               regression coefficients, respectively, onto the orthogonal
%               complement of A.
%
%   'lags'      Scalar or vector of nonnegative integers indicating the
%               number q of lagged differences in the VEC(q) model of y(t).
%               The default value is 0.
%
%               Lagging and differencing a time series reduce the sample
%               size. Absent any presample values, if y(t) is defined for
%               t = 1:N, then the lagged series y(t-k) is defined for
%               t = k+1:N. Differencing reduces the time base to k+2:N.
%               With q lagged differences, the common time base is q+2:N
%               and the effective sample size is T = N-(q+1).
%
%   'alpha'     Scalar or vector of nominal significance levels for the
%               tests. Values must be greater than zero and less than one.
%               The default value is 0.05.
%
%   Single-element values for inputs are expanded to the length of any
%   vector value (the number of tests). Vector values must have equal
%   length. If any value is a row vector, all outputs are row vectors.
%
% Output Arguments:
%
%   h - Vector of Boolean decisions for the tests, with length equal to the
%       number of tests. Values of h equal to 1 (true) indicate rejection
%       of the null that the constraints hold in favor of the alternative
%       that they do not. Values of h equal to 0 (false) indicate a failure
%       to reject the null.
%
%   pValue - Vector of right-tail probabilities of the test statistics,
%       with length equal to the number of tests.
%
%   stat - Vector of test statistics, with length equal to the number of
%       tests. Statistics are likelihood ratios determined by the test.
%
%   cValue - Critical values for right-tail probabilities, with length
%       equal to the number of tests. The asymptotic distributions of the
%       test statistics are chi-square, with degree-of-freedom parameter
%       determined by the test.
%
%   mles - Structure of maximum likelihood estimates associated with the
%       VEC(q) model of y(t), subject to the constraints. Each structure
%       has the following fields:
%
%       paramNames  Cell vector of parameter names, of the form:
%
%                   	{'A','B','B1',...,'Bq','c0','d0','c1','d1'}'.
%
%                   Elements depend on the values of 'lags' and 'model'.
%
%       paramVals   Structure of parameter estimates with field names
%                   corresponding to the parameter names in paramNames.
%
%       res         T-by-numDims matrix of residuals, where T is the
%                   effective sample size, obtained by fitting the VEC(q)
%                   model of y(t) to the input data.
%
%       EstCov      Estimated covariance Q of the innovations process e(t).
%
%       rLL         Restricted loglikelihood of Y under the null.
%
%       uLL         Unrestricted loglikelihood of Y under the alternative.
%
%       dof         Degrees of freedom of the asymptotic chi-square
%                   distribution of the test statistic.
%
% Notes:
%
%   o The parameters A and B in the reduced-rank VEC(q) model are not
%     uniquely identified. JCONTEST identifies B using the methods in [3],
%     depending on the test.
%
%   o When constructing constraints, interpret the rows and columns of the
%     numDims-by-r matrices A and B as follows:
%
%     - Row i of A contains the adjustment speeds of variable y_i to
%       disequilibrium in each of the r cointegrating relations.
%
%     - Column j of A contains the adjustment speeds of each of the numDims
%       variables to disequillibrium in the jth cointegrating relation.
%
%     - Row i of B contains the coefficients of variable y_i in each of the
%       r cointegrating relations.
%
%     - Column j of B contains the coefficients of each of the numDims
%       variables in the jth cointegrating relation.
%
%   o Tests on B answer questions about the space of cointegrating
%     relations. Tests on A answer questions about common driving forces in
%     the system. For example, an all-zero row in A indicates a variable
%     that is weakly exogenous with respect to the coefficients in B. Such
%     a variable may affect other variables, but it does not adjust to
%     disequillibrium in the cointegrating relations. Similarly, a standard
%     unit vector column in A indicates a variable that is exclusively
%     adjusting to disequillibrium in a particular cointegrating relation.
%
%   o Constraints matrices R satisfying R'*A = 0 or R'*B = 0 are equivalent
%     to A = H*phi or B = H*phi, where H is the orthogonal complement of R
%     (null(R')) and phi is a vector of free parameters.
%
%   o JCONTEST compares finite-sample statistics to asymptotic critical
%     values, and tests can show significant size distortions for small
%     samples. See [2]. Larger samples lead to more reliable inferences.
%
%   o To convert VEC(q) model parameters in the mles output to vector
%     autoregressive (VAR) model parameters, use the utility VECTOVAR.
%
% Example:
%
%   % Data on Australian and U.S. prices:
% 
%   load Data_JAustralian
%   p1 = Dataset.PAU; % Log Australian Consumer Price Index
%   p2 = Dataset.PUS; % Log U.S. Consumer Price Index
%   s12 = Dataset.EXCH; % Log AUD/USD Exchange Rate
%   Y = [p1 p2 s12];
%   plot(dates,Y)
%   datetick('x','yyyy')
%   legend(series(1:3),'Location','Best')
%   grid on
% 
%   % Pretest the individual series for stationarity:
% 
%   [h0,pValue0] = jcontest(Y,1,'BVec',{[1 0 0]',[0 1 0]',[0 0 1]'})
% 
%   % Test for cointegration:
% 
%   [h1,pValue1] = jcitest(Y);
% 
%   % Test for purchasing power parity (p1 = p2 + s12):
% 
%   [h2,pValue2] = jcontest(Y,1,'BCon',[1 -1 -1]')
%
% References:
% 
%   [1] Hamilton, J. D. Time Series Analysis. Princeton, NJ: Princeton
%       University Press, 1994.
%
%   [2] Haug, A. "Testing Linear Restrictions on Cointegrating Vectors:
%       Sizes and Powers of Wald Tests in Finite Samples." Econometric
%       Theory. v. 18, 2002, pp. 505-524.
%
%   [3] Johansen, S. Likelihood-Based Inference in Cointegrated Vector
%       Autoregressive Models. Oxford: Oxford University Press, 1995.
%
%   [4] Juselius, K. The Cointegrated VAR Model. Oxford: Oxford University
%       Press, 2006.
%
%   [5] Morin, N. "Likelihood Ratio Tests on Cointegrating Vectors,
%       Disequilibrium Adjustment Vectors, and their Orthogonal
%       Complements." European Journal of Pure and Applied Mathematics.
%       v. 3, 2010, pp. 541-571.
%
% See also JCITEST.

% Copyright 2011 The MathWorks, Inc.
% $Revision: 1.1.6.7 $ $Date: 2012/08/21 00:03:36 $

% Parse inputs and set defaults:

parseObj = inputParser;
parseObj.addRequired('Y',@YCheck);
parseObj.addRequired('r',@rCheck);
parseObj.addRequired('test',@testCheck);
parseObj.addRequired('Cons',@consCheck);
parseObj.addParamValue('model','H1',@modelCheck);
parseObj.addParamValue('lags',0,@lagsCheck);
parseObj.addParamValue('alpha',0.05,@alphaCheck);

parseObj.parse(Y,r,test,Cons,varargin{:});

Y = parseObj.Results.Y;
r = parseObj.Results.r;
test = lower(parseObj.Results.test);
Cons = parseObj.Results.Cons;
model = upper(parseObj.Results.model);
lags = parseObj.Results.lags;
alpha = parseObj.Results.alpha;

% Check inputs for commensurate lengths, expand single-element values, and
% convert all variables to columns:

[numTests,rowOutput,r,test,Cons,model,lags,alpha] = ...
    sizeCheck(r,test,Cons,model,lags,alpha);

% Remove rows of Y with missing values:

Y(any(isnan(Y),2),:) = [];
[numObs,numDims] = size(Y);

% Check if any lags are too big:

bigLags = find(numObs < lags+2);
if ~isempty(bigLags)

    error(message('econ:jcontest:ZeroEffectiveSampleSize', bigLags))

end

% Check if rank specifications are out of range:

if any(r < 1) || any(r > numDims-1)
    
    error(message('econ:jcontest:RankOutOfRange'))
      
end

% Check the size of the constraints matrices before running the tests:

for testNum = 1:numTests
    
    testCons = Cons{testNum};
    testType = test{testNum};
    testRank = r(testNum);
    
    [m,n] = size(testCons);
    
    switch testType
        
        case 'acon'
                  
            if m ~= numDims

                error(message('econ:jcontest:AConRowDimension', testNum))

            end
            
            if n > numDims-testRank

                error(message('econ:jcontest:AConColumnDimension', testNum))

            end
            
        case 'avec'
            
            if m ~= numDims

                error(message('econ:jcontest:AVecRowDimension', testNum))

            end
            
          	if n > testRank

                error(message('econ:jcontest:AVecColumnDimension', testNum))

            end
            
        case 'bcon'
            
            testModel = model{testNum};
                
            if any(strcmp(testModel,{'H*','H1*'}))
        
                numDims1 = numDims+1; % Add restricted deterministic term
            
            else
                
                numDims1 = numDims;
                
            end
            
            if m ~= numDims1

                error(message('econ:jcontest:BConRowDimension', testNum))

            end
            
            n0 = sum(all(testCons(1:m-1,:) == 0) & (testCons(m,:) ~= 0));
            EndCons = max(n-n0,0); % Count only endogenous constraints
            
            if EndCons > numDims-testRank

                error(message('econ:jcontest:BConColumnDimension', testNum))

            end
            
        case 'bvec'
            
            testModel = model{testNum};
                
            if any(strcmp(testModel,{'H*','H1*'}))
        
                numDims1 = numDims+1; % Add restricted deterministic term
            
            else
                
                numDims1 = numDims;
                
            end
            
            if m ~= numDims1

                error(message('econ:jcontest:BVecRowDimension', testNum))

            end
            
            if n > testRank

                error(message('econ:jcontest:BVecColumnDimension', testNum))

            end
            
    end
    
end

% Preallocate output variables:

switch nargout
    
    case {0,1}
        
        needPValue = false;
        needMLEs = false;
        
        h = false(numTests,1);
        
    case 2
        
        needPValue = true;
        needMLEs = false;
        
        h = false(numTests,1);
        pValue = NaN(numTests,1);
        
    case 3
        
        needPValue = true;
        needMLEs = false;
        
        h = false(numTests,1);
        pValue = NaN(numTests,1);
        stat = pValue;
        
    case 4
        
        needPValue = true;
        needMLEs = false;
        
        h = false(numTests,1);
        pValue = NaN(numTests,1);
        stat = pValue;
        cValue = pValue;
        
    case 5
        
        needPValue = true;
        needMLEs = true;
        
        h = false(numTests,1);
        pValue = NaN(numTests,1);
        stat = pValue;
        cValue = pValue;
        MLEFields = {'paramNames','paramVals','res','EstCov','rLL','uLL','dof'};
        mles = cell2struct(cell(length(MLEFields),numTests),MLEFields,1);

end

% Initialize loop variables:

lastR = [];
lastType = [];
lastCons = [];
lastModel = [];
lastLags = [];

% Run the tests:

for testNum = 1:numTests
    
    testRank = r(testNum);
    testType = test{testNum};
    testCons = Cons{testNum};
    testModel = model{testNum};
    testLags = lags(testNum);
    testAlpha = alpha(testNum);
    
    tBase = (testLags+2):numObs; % Commensurate time base, all lags
    testT = length(tBase); % Effective sample size
    
    [numDims1,numCons] = size(testCons); % Count all constraints for dof
  
    % Check to see if statistics are unchanged from last test:
    
    sameStat = isequal(testRank,lastR) && ...
               isequal(testType,lastType) && ...
               isequal(testCons,lastCons) && ...
               isequal(testModel,lastModel) && ...
               isequal(testLags,lastLags);
    
    if ~sameStat % Recompute the statistic only if it changes
        
        % The general procedure for computing and testing the chi-square
        % statistics is outlined below. Components of the procedure switch
        % on testType to implement the appropriate type of test. See:
        %
        % 'ACon': [3] Theorem 8.2, [4] Section 11.1
        % 'AVec': [3] Theorem 8.4, [4] Section 11.3
        % 'BCon': [3] Theorem 7.2, [4] Section 10.2
        % 'BVec': [3] Theorem 7.3, [4] Section 10.3
        
        % Find the orthogonal complement of the constraints matrix:
        
        H = null(testCons');
        HBar = H/(H'*H);
        
        % Perform the concentrating regressions and return the product
        % moment matrices of the residuals. Also return LY, DY, and DLY for
        % use in computing MLEs:
        
        [moMa,LY,DY,DLY] = conVEC;
        
        % Extract the product moment matrices common to all test types:
        
        S00 = moMa{1};
        S01 = moMa{2};
        S10 = moMa{3};
        S11 = moMa{4};
        
        % Estimate the eigenvectors and eigenvalues:
        
        [V,lambda] = eigProb;
                
        % Normalize the eigenvectors:
        
        V = normalizeV;
        
        % Compute the statistic:
        
        [testStat,logLambda] = getStat;
        
        % Compute the degrees of freedom:
        
        dof = getDof;
        
        % Compute the p-value, if requested:
        
        if needPValue
            
            testPValue = 1-chi2cdf(testStat,dof);
            
        end
               
    end
    
    % Test the statistic:
    
    testCValue = chi2inv(1-testAlpha,dof);
    testH = (testStat > testCValue); % Right-tailed test
    
    % Compute mles, if requested:
    
    if needMLEs
        
        testMLEs = getMLEs;
        
    end

    % Add the test results to the outputs:
    
    switch nargout
    
        case {0,1}
        
            h(testNum) = testH;
        
        case 2
        
            h(testNum) = testH;
            pValue(testNum) = testPValue;
        
        case 3
        
            h(testNum) = testH;
            pValue(testNum) = testPValue;
            stat(testNum) = testStat;
        
        case 4
        
            h(testNum) = testH;
            pValue(testNum) = testPValue;
            stat(testNum) = testStat;
            cValue(testNum) = testCValue;
        
        case 5
        
            h(testNum) = testH;
            pValue(testNum) = testPValue;
            stat(testNum) = testStat;
            cValue(testNum) = testCValue;
            mles(testNum) = testMLEs;
            
    end
    
    % Save values to check for changes, next loop:
    
    lastR = testRank;
    lastType = testType;
    lastCons = testCons;
    lastModel = testModel;
    lastLags = testLags;

end

% Display outputs as row vectors if any parameter value is a row vector:

if rowOutput
    
    switch nargout
        
        case {0,1}

            h = h';

        case 2
            
            h = h';
            pValue = pValue';
            
        case 3
            
            h = h';
            pValue = pValue';
            stat = stat';
            
        case 4
            
            h = h';
            pValue = pValue';
            stat = stat';
            cValue = cValue';
            
        case 5
            
            h = h';
            pValue = pValue';
            stat = stat';
            cValue = cValue';
            mles = mles';
        
    end
    
end

%-------------------------------------------------------------------------
% Check input Y
function OK = YCheck(Y)
            
    if isempty(Y)
        
        error(message('econ:jcontest:DataUnspecified'))
          
    elseif isvector(Y)
        
        error(message('econ:jcontest:DataIsVector'))
          
                    
    elseif ~isnumeric(Y)
        
        error(message('econ:jcontest:DataNonNumeric'))
          
    else
        
        OK = true;
        
    end
    
end % YCHECK

%-------------------------------------------------------------------------
% Check input r
function OK = rCheck(r)
            
    if isempty(r)
        
        error(message('econ:jcontest:RankUnspecified'))
          
    elseif ~isvector(r)
        
        error(message('econ:jcontest:RankNonVector'))
          
                    
    elseif ~isnumeric(r)
        
        error(message('econ:jcontest:RankNonNumeric'))
          
    elseif any(mod(r,1) ~= 0)
        
        error(message('econ:jcontest:RankNonInteger'))
          
    else
        
        OK = true;
        
    end
    
end % RCHECK

%-------------------------------------------------------------------------
% Check input test
function OK = testCheck(test)
            
    if ~isvector(test)
        
        error(message('econ:jcontest:TestNonVector'))
          
    elseif isnumeric(test) || (iscell(test) && any(cellfun(@isnumeric,test)))
        
        error(message('econ:jcontest:TestNumeric'))
    
    elseif ~all(ismember(lower(test),{'acon','avec','bcon','bvec'}))
        
        error(message('econ:jcontest:TestInvalid'))
          
    else
        
        OK = true;
        
    end
    
end % TESTCHECK

%-------------------------------------------------------------------------
% Check input Cons
function OK = consCheck(Cons)
            
    if isempty(Cons)
        
        error(message('econ:jcontest:ConsUnspecified'))
                    
    elseif (~iscell(Cons) && ~isnumeric(Cons)) || (iscell(Cons) && ~all(cellfun(@isnumeric,Cons)))
        
        error(message('econ:jcontest:ConsNonNumeric'))
          
    else
        
        OK = true;
        
    end
    
end % CONSCHECK

%-------------------------------------------------------------------------
% Check value of 'model' parameter
function OK = modelCheck(model)
    
    if ~isvector(model)
        
        error(message('econ:jcontest:ModelNonVector'))
          
    elseif isnumeric(model) || (iscell(model) && any(cellfun(@isnumeric,model)))
        
        error(message('econ:jcontest:ModelNumeric'))
    
    elseif ~all(ismember(upper(model),{'H2','H1*','H1','H*','H'}))
        
        error(message('econ:jcontest:ModelInvalid'))
          
    else
        
        OK = true;
        
    end
    
 end % MODELCHECK

%-------------------------------------------------------------------------
% Check value of 'lags' parameter
function OK = lagsCheck(lags)
    
    if ~isvector(lags)
        
        error(message('econ:jcontest:LagsNonVector'))
          
    elseif ~isnumeric(lags)
        
        error(message('econ:jcontest:LagsNonNumeric'))
          
    elseif any(mod(lags,1) ~= 0) || any(lags < 0)
        
        error(message('econ:jcontest:LagsOutOfRange'))
          
    else
        
        OK = true;
        
    end
    
end % LAGSCHECK

%-------------------------------------------------------------------------
% Check value of 'alpha' parameter
function OK = alphaCheck(alpha)
    
    if ~isnumeric(alpha)
        
        error(message('econ:jcontest:AlphaNonNumeric'))
          
    elseif ~isvector(alpha)
        
        error(message('econ:jcontest:AlphaNonVector'))
          
    elseif any(alpha <= 0) || any(alpha >= 1)
        
        error(message('econ:jcontest:AlphaOutOfRange'))
          
    else
        
        OK = true;
        
    end
    
end % ALPHACHECK

%-------------------------------------------------------------------------
% Check parameter values for commensurate lengths, expand single-element
% values, and convert all variables to columns
function [numTests,rowOutput,varargout] = sizeCheck(varargin)

% Initialize outputs:

numTests = 1;
rowOutput = false;

% Determine vector lengths, number of tests:

for i = 1:nargin
    
    ivar = varargin{i};
    iname = inputname(i);
    
    if (isnumeric(ivar) && ~strcmp(iname,'Cons')) || iscell(ivar)
        paramLength.(iname) = length(ivar);
        if ~isscalar(ivar)
        	rowOutput = rowOutput || (size(ivar,1) == 1);
        end    
    else        
        paramLength.(iname) = 1;   % Single string or matrix
        varargin{i} = varargin(i); % Convert to cell        
    end
    
    numTests = max(numTests,paramLength.(iname));
    
end

% Check for commensurate vector lengths:

for i = 1:(nargin-1)
    iname = inputname(i);
    for j = (i+1):nargin
        jname = inputname(j);
        if (paramLength.(iname) > 1) && (paramLength.(jname) > 1) ...
            && (paramLength.(iname) ~= paramLength.(jname))
        
            error(message('econ:jcontest:ParameterSizeMismatch', iname, jname))
              
        end        
    end
end

% Expand scalars and single strings:

for i = 1:nargin
    
    ivar = varargin{i};
    if paramLength.(inputname(i)) == 1
        varargout{i} = ivar(ones(numTests,1)); % Expand to column output
    else
        varargout{i} = ivar(:);  % Column output
    end
    
end

end % SIZECHECK

%-------------------------------------------------------------------------
% Concentrate the VEC model
function [moMa,LY,DY,DLY] = conVEC
    
YLags = lagmatrix(Y,0:(testLags+1)); % Y(t-k) on observed time base
LY = YLags(tBase,(numDims+1):2*numDims); % Y(t-1) on commensurate time base

% Form multidimensional differences so that the kth numDims-wide block of
% columns in DelatYLags contains (1-L)Y(t-k+1):

DeltaYLags = zeros(testT,(testLags+1)*numDims);
for k = 1:(testLags+1)
    DeltaYLags(:,((k-1)*numDims+1):k*numDims) = ...
               YLags(tBase,((k-1)*numDims+1):k*numDims) ...
             - YLags(tBase,(k*numDims+1):(k+1)*numDims);
end

DY = DeltaYLags(:,1:numDims); % (1-L)Y(t)
DLY = DeltaYLags(:,(numDims+1):end); % [(1-L)Y(t-1), ..., (1-L)Y(t-q)]

% Express the VEC(q) model in the form Z0 = A*B'*Z1 + PSI*Z2 + e(t):

switch testModel

    case 'H2'

        Z0 = DY;
        Z1 = LY;
        Z2 = DLY;

    case 'H1*'

        Z0 = DY;
        Z1 = [LY,ones(testT,1)];  % Restricted constant
        Z2 = DLY;

    case 'H1'

        Z0 = DY;
        Z1 = LY;
        Z2 = [DLY,ones(testT,1)];  % Unrestricted constant

    case 'H*'

        Z0 = DY;
        Z1 = [LY,(1:testT)'];  % Restricted linear term
        Z2 = [DLY,ones(testT,1)]; % Unrestricted constant

    case 'H'

        Z0 = DY;
        Z1 = LY;
        Z2 = [DLY,ones(testT,1),(1:testT)']; % Unrestricted constant, linear term

end

% Compute the residuals of the concentrating regressions:

R0 = Z0-Z2*(Z2\Z0); % Z0 on Z2 residuals
R1 = Z1-Z2*(Z2\Z1); % Z1 on Z2 residuals

% Compute the product moment matrices common to all test types:

S00 = (R0'*R0)/testT;
S01 = (R0'*R1)/testT;
S10 = (R1'*R0)/testT;
S11 = (R1'*R1)/testT;

% Compute additional product moment matrices, as necessary:

switch testType
    
    case 'acon'
        
        X00 = testCons'*S00*testCons;
        XX00 = testCons*(X00\testCons');

        S00a = S00-S00*XX00*S00;
        S01a = S01-S00*XX00*S01;
        S10a = S10-S10*XX00*S00;
        S11a = S11-S10*XX00*S01;
        
        moMa = {S00,S01,S10,S11,S00a,S01a,S10a,S11a};
        
    case 'avec'
        
        moMa = cell(1,10); % Preallocate for later additions
        moMa(1:6) = {S00,S01,S10,S11,R0,R1};
        
    case 'bcon'

        moMa = {S00,S01,S10,S11};
        
    case 'bvec'
        
        X11 = testCons'*S11*testCons;
        XX11 = testCons*(X11\testCons');
        
        S00b = S00-S01*XX11*S10;
        S01b = S01-S01*XX11*S11;
        S10b = S10-S11*XX11*S10;
        S11b = S11-S11*XX11*S11;
        
        moMa = {S00,S01,S10,S11,S00b,S01b,S10b,S11b,X11,XX11};
        
end
    
end % CONVEC

%-------------------------------------------------------------------------
% Estimate the eigenvectors and eigenvalues
function  [V,lambda] = eigProb
    
% Unconstrained:

[~,DHat] = eig(S10*(S00\S01),S11);
lambdaHat = sort(diag(DHat),'descend');

% Constrained:

switch testType
    
    case 'acon'
        
        S00a = moMa{5};
        S01a = moMa{6};
        S10a = moMa{7};
        S11a = moMa{8};
        
        HXa = H*((H'*S00a*H)\H');
        
        [VStar,DStar] = eig(S10a*HXa*S01a,S11a);
        [lambdaStar,sortIdx] = sort(diag(DStar),'descend');        
        V = VStar(:,sortIdx);
        
        lambda = [lambdaHat(1:testRank),lambdaStar(1:testRank)];
        
    case 'avec'
        
        HX = H*((H'*S00*H)\H');
        
        [VStar,DStar] = eig(S10*HX*S01,S11);
        [lambdaStar,sortIdx] = sort(diag(DStar),'descend');
        V = VStar(:,sortIdx);        
        
        U = V'*S11*V; % V diagonalizes S11
        V = bsxfun(@rdivide,V,sqrt(diag(U))'); % Normalize V
        
        RC = testRank-numCons;
        Phi = V(:,1:RC);
        Psi = H'*S01*Phi;
        
        R0 = moMa{5};
        R1 = moMa{6};
        Rk = R0*H-R1*Phi*Psi';
        
        S0k = (R0'*Rk)/testT;
        S1k = (R1'*Rk)/testT;
        Sk0 = (Rk'*R0)/testT;
        Sk1 = (Rk'*R1)/testT;
        Skk = (Rk'*Rk)/testT;
        
        S00k = S00-S0k*(Skk\Sk0);
        S01k = S01-S0k*(Skk\Sk1);
        S10k = S10-S1k*(Skk\Sk0);
        S11k = S11-S1k*(Skk\Sk1);
        
        moMa{7} = S00k;
        moMa{8} = S10k;
        moMa{9} = S11k;
        
        X00k = testCons'*S00k*testCons;
        XX00k = testCons*(X00k\testCons');
        dStar = eig(S10k*XX00k*S01k,S11k);
        rhoStar = sort(dStar,'descend');
        
        lambda = zeros(testRank,3);
        lambda(:,1) = lambdaHat(1:testRank);
        lambda(1:RC,2) = lambdaStar(1:RC);
        lambda(1:numCons,3) = rhoStar(1:numCons);
        
    case 'bcon'
        
        [VStar,DStar] = eig(H'*S10*(S00\S01)*H,H'*S11*H);
        [lambdaStar,sortIdx] = sort(diag(DStar),'descend');
        V = VStar(:,sortIdx);
        
        lambda = [lambdaHat(1:testRank),lambdaStar(1:testRank)];
        
    case 'bvec'
        
        S00b = moMa{5};
        S01b = moMa{6};
        S10b = moMa{7};
        S11b = moMa{8};
        XX11 = moMa{10};
        
        [VStar,DStar] = eig(H'*S10b*(S00b\S01b)*H,H'*S11b*H);
        [lambdaStar,sortIdx] = sort(diag(DStar),'descend');
        V = VStar(:,sortIdx);
        
        dStar = eig(S01*XX11*S10,S00);
        rhoStar = sort(dStar,'descend');
        
        RC = testRank-numCons;
        lambda = zeros(testRank,3);
        lambda(1:testRank,1) = lambdaHat(1:testRank);
        lambda(1:RC,2) = lambdaStar(1:RC);
        lambda(1:numCons,3) = rhoStar(1:numCons);
        
end    
    
end % EIGPROB

%-------------------------------------------------------------------------
% Normalize the eigenvectors
function W = normalizeV

switch testType
    
    case 'acon'
        
        S11a = moMa{8};
        
        U = V'*S11a*V; % V diagonalizes S11a
        W = bsxfun(@rdivide,V,sqrt(diag(U))');
        
    case 'avec'
        
        W = V; % Normalized in eigProb
        
    case 'bcon'
        
        U = V'*H'*S11*H*V; % H*V diagonalizes S11
        W = bsxfun(@rdivide,V,sqrt(diag(U))');
        
    case 'bvec'
        
        S11b = moMa{8};
        
        U = V'*H'*S11b*H*V; % H*V diagonalizes S11b
        W = bsxfun(@rdivide,V,sqrt(diag(U))');
        
end    
    
end % NORMALIZEV

%-------------------------------------------------------------------------
% Compute the statistic
function [testStat,logLambda] = getStat

% Form summands and compute statistic:

logLambdaHat = log(abs(1-lambda(:,1)));
logLambdaStar = log(abs(1-lambda(:,2)));

switch testType
    
    case {'acon','bcon'}
        
        logLambdaRatio = logLambdaStar-logLambdaHat;
        testStat = testT*sum(logLambdaRatio(1:testRank));
        
        logLambda = [logLambdaHat,logLambdaStar];
        
    case 'avec'
        
        logRhoStar = log(abs(1-lambda(:,3)));
        
        S00k = moMa{7};
        D1 = det(testCons'*S00k*testCons);
        D2 = det(H'*S00*H);
        D3 = det(testCons'*testCons);
        D4 = det(H'*H);
        logDX = log(D1)+log(D2)-log(D3)-log(D4);
        moMa{10} = logDX;
        
        testStat = testT*(logDX-log(det(S00)) + ...
                          sum(logLambdaStar(1:testRank-numCons)) + ...
                          sum(logRhoStar(1:numCons)) - ...
                          sum(logLambdaHat(1:testRank)));
        
        logLambda = [logLambdaHat,logLambdaStar,logRhoStar];
        
    case 'bvec'
        
        logRhoStar = log(abs(1-lambda(:,3)));
        
        testStat = testT*(sum(logLambdaStar(1:testRank-numCons)) + ...
                          sum(logRhoStar(1:numCons)) - ...
                          sum(logLambdaHat(1:testRank)));
                      
        logLambda = [logLambdaHat,logLambdaStar,logRhoStar];
        
end

end % GETSTAT

%-------------------------------------------------------------------------
% Get degrees of freedom:
function dof = getDof
    
switch testType
    
    case {'acon','bcon'}
        
        dof = testRank*numCons;
        
    case {'avec','bvec'}
        
        dof = (numDims1-testRank)*numCons;
        
end

end % GETDOF

%-------------------------------------------------------------------------
% Compute MLEs of the VEC model
function testMLEs = getMLEs
    
switch testType
    
    case 'acon'
        
        S01a = moMa{6};
        Q = V(:,1:testRank);
        A = HBar*H'*S01a*Q;
        B = Q(1:numDims,:);
        
    case 'avec'
        
        S10k = moMa{8};
        S11k = moMa{9};
        aBar = testCons/(testCons'*testCons);
        B1 = (S11k\S10k)*aBar;        
        B2 = V(:,1:testRank-numCons);
        Q = [B1,B2];
        tau = HBar*H'*S01*B2;
        A = [testCons,tau];
        B = Q(1:numDims,:);
        
    case 'bcon'

        Q = H*V(:,1:testRank);
        A = S01*Q;
        B = Q(1:numDims,:);

    case 'bvec'
        
        S01b = moMa{6};
        X11 = moMa{9};
        Q2 = H*V(:,1:testRank-numCons);
        Q = [testCons,Q2];
        A = [(S01*testCons-S01b*(Q2*Q2')*S11*testCons)/X11,S01b*Q2];
        B = Q(1:numDims,:);
        
        
end

switch testModel

        case 'H2'
            
            detNames = cell(0,1);
            W = DY-LY*B*A';
            X = DLY;
            P = (X\W)'; % [B1, ..., Bq]
            detVals = cell(0,1);
            [paramNames,paramVals] = assembleParams;
            
        case 'H1*'
            
            detNames = {'c0'};
            c0 = Q(end,:)';
            W = DY-(LY*B+repmat(c0',testT,1))*A';
            X = DLY;
            P = (X\W)'; % [B1, ..., Bq]
            detVals = {c0};
            [paramNames,paramVals] = assembleParams;
            
        case 'H1'
            
            detNames = {'c0';'c1'};
            W = DY-LY*B*A';
            X = [DLY,ones(testT,1)];
            P = (X\W)'; % [B1, ..., Bq, c]            
            c = P(:,end);
            c0 = A\c; % Project c on A
            c1 = c-A*c0; % Orthogonal component of c
            detVals = {c0;c1};
            [paramNames,paramVals] = assembleParams;
            
        case 'H*'
            
            detNames = {'c0';'d0';'c1'};
            d0 = Q(end,:)';
            W = DY-(LY*B+(1:testT)'*d0')*A';
            X = [DLY,ones(testT,1)];
            P = (X\W)'; % [B1, ..., Bq, c]
            c = P(:,end);
            c0 = A\c; % Project c on A
            c1 = c-A*c0; % Orthogonal component of c
            detVals = {c0;d0;c1};
            [paramNames,paramVals] = assembleParams;
            
        case 'H'
            
            detNames = {'c0';'d0';'c1';'d1'};
            W = DY-LY*B*A';
            X = [DLY,ones(testT,1),(1:testT)'];
            P = (X\W)'; % [B1, ..., Bq, c, d]
            c = P(:,end-1);
            c0 = A\c; % Project c on A
            c1 = c-A*c0; % Orthogonal component of c
            d = P(:,end);
            d0 = A\d; % Project d on A
            d1 = d-A*d0; % Orthogonal component of d
            detVals = {c0;d0;c1;d1};
            [paramNames,paramVals] = assembleParams;
            
end

testMLEs.paramNames = paramNames;    
testMLEs.paramVals = paramVals;
res = W-X*P';
testMLEs.res = res;
testMLEs.EstCov = cov(res); % This estimates the MLE, which depends on test
testMLEs.rLL = getRLL;
testMLEs.uLL = -(testT*numDims/2)*(log(2*pi)+1) ...
               -(testT/2)*(log(det(S00))+sum(logLambda(1:testRank,1)));
testMLEs.dof = dof;
    
    %----------------------------------------------------------------------
    % Assemble parameter names/values
    function [paramNames,paramVals] = assembleParams

    if testLags > 0
        lagNames = strcat({'B'},num2str((1:(testLags))','%-d'));
    else
        lagNames = cell(0,1);
    end

    paramNames = cat(1,'A','B',lagNames,detNames);
    
    lagVals = cell(testLags,1);
    for lag = 1:testLags        
        lagVals{lag} = P(:,(((lag-1)*numDims)+1):lag*numDims);       
    end

    paramVals0 = cat(1,{A},{B},lagVals,detVals);
    
    % Convert to structure with dynamic field names:
    
    for param = 1:length(paramNames)
        paramVals.(paramNames{param}) = paramVals0{param};
    end

    end % ASSEMBLEPARAMS
    
    %----------------------------------------------------------------------
    % Compute restricted loglikelihood under constraints on H(r)
    function rLL = getRLL
        
    switch testType
    
    	case {'acon','bcon'}
        
            rLL = -(testT*numDims/2)*(log(2*pi)+1)...
                  -(testT/2)*(log(det(S00))+sum(logLambda(1:testRank,2)));
              
        case 'avec'
            
            logDX = moMa{10};
            rLL = -(testT*numDims/2)*(log(2*pi)+1)...
                  -(testT/2)*(logDX + ...
                              sum(logLambda(1:testRank-numCons,2)) + ...
                              sum(logLambda(1:numCons,3)));

     	case 'bvec'
        
            rLL = -(testT*numDims/2)*(log(2*pi)+1)...
                  -(testT/2)*(log(det(S00))+ ...
                              sum(logLambda(1:testRank-numCons,2))+ ...
                              sum(logLambda(1:numCons,3)));
        
    end

    end % GETRLL

end % GETMLES

%-------------------------------------------------------------------------

end % JCONTEST