% -------------------------------------------------------------------------- */
%                                  VAR.G
%     general purpose proc for doing var's of cointegrated systems in error
%     correction form.
% 
%     {b,sigma,r,Dxir,xir,statcom,e} = var(x,a,lags,horiz,names,bigres,prnt)
% 
%     Inputs:
%     x     = series, from 0 to T, in levels.  shocks are orthogonalized in the
%              order given here.
%     a     = cointegrating vector. x*a is
%              stationary and is added to the right hand side of the VAR
%     lags  = number of lags of differences to include
%     horiz = how far out to do impulse response function
%     names = vector of names of the cointegrating vectors and then variables
%             no more than 3 letters as (t-x) are added in the program.
%             e.g. names = Y-C, y, c.
%     bigres = 1 extra stuff; 0= bare bones for monte carlo.
%     print = 1 to print out results, 0 to not print.
% 
%     Outputs:
%     b: OLS regression coefficients:
%                          Dx1(t)   Dx2(t)  Dx3(t) ...
%              const
%              ci1(t-1)
%              ci2(t-1)
%              ..
%              Dx1(t-1)
%              Dx1(t-2)
%              ..
%              Dx1(t-lags)
%              Dx2(t-1)
%              ...
% 
%      r: choleski decomposition of sigma; sigma = r'r
%      Dxir: impulse response of Dx.
%                           response of
%                 Dx1 Dx2 Dx3 ..    Dx1 Dx2 Dx3 ...
%                         to unit shock in
%                 Dx1 Dx1 Dx1       Dx2 Dx2 Dx2 ...
%           t
%           t+1
%           t+2
%           ...
%           t+horiz
% 
%       statcom: Beveridge Nelson stationary component
%       e: error series
% 
% 
% 
%     The proc runs a var of the form
% 
%     A(L) Dx(t) =  B * (x(t-1)*a) + e.
% 
%     by OLS. It uses the choleski decomposition
% 
%     var(ee') = sigma = rr'
% 
%     so
% 
%     e = rv  E(vv') = I
% 
%     the final form of the VAR is then
% 
%        -1                 -1
%     (r)  A(L) Dx(t) =  (r)   B * (x(t-1)*a) + v
% 
%     It finds impulse responses by simulating the OLS system (not the constants)
%     starting from 0.
% 
% 
% ------------------------------------------------------------------------- */

function [b,sigma,r,Dxir,xir,statcom,e] = varEC(x,a,lags,horiz,names,bigres,prnt)
     
statcom = 0;

% checks 
    if rows(a) ~= cols(x);
        disp('proc var: cointegrating vector not conformable to data');
    end;
    if rows(names) ~= (cols(a)+cols(x));
        disp('proc var: names not conformable to data, ci vectors');
    end;

% ------------------------------------------------------------ 
%             set up right hand variables                      
% Rhv = const ci(t-1) .. x1(t-1) x1(t-2) .. x2(t-1) x2(t-2) .. 
% ------------------------------------------------------------ 

    Dx = x(2:rows(x),:)-x(1:rows(x)-1,:);
    xl = x(1:rows(x)-1,:);                       % note x is LAGGED levels
    T = rows(xl);
    nseries = cols(xl);


    Rhv = [ones(T-lags,1) (xl(1+lags:T,:)*a)];    % constant, lagged ci vec*/
    if lags > 0;                                  % then lags of Dx */
        i = 1;
        while i <= cols(xl);
            j = 1;
            while j <= lags;
                Rhv = [Rhv Dx(1+lags-j:T-j,i)];
                j = j+1;
            end;
            i = i+1;
        end;
    end;

% ----------------------------- 
% run regression of Dx's on Rhv 
% ----------------------------- 

   b = Rhv\Dx(1+lags:T,:);                 % b is a matrix.  each column gives ols of that lhv on all Rhv's  
   e = Dx(1+lags:T,:)-Rhv*b;
   sigma = e'*e/(T-lags);
   seb = (diag(inv(Rhv'*Rhv))*diag(sigma)').^(1/2);
   R2 = 1 - (diag(sigma)')./(std(Dx(1+lags:T,:)).^2);
   fstat = (R2./(1-R2)).*(T-cols(xl)-1)/cols(xl);
   pvf = 0*fstat;
   i = 1;
   while i <= rows(fstat);
        pvf(i) = 1-fcdf(fstat(i),cols(xl),T-cols(xl)-1);
        i = i+1;
   end;


% --------------------------------- 
%  A(L)Dx(t) = mu + G(a'x(t-1)) + e 
%                representation.    
%  AR is A0~A1~A2...                
% --------------------------------- 
  if bigres;
      AR = eye(cols(Dx));
      offst = cols(a) + 1;
      i = 1;
     while i <= lags;
        j = 1;
        while j <= cols(Dx);
            AR = [AR (-b(offst+i+(j-1)*lags,:)')];
            j = j+1;
        end;
        i = i+1;
     end;

      G =  b(2:cols(a)+1,:)';
      mu = b(1,:)';
  end;
% ----------------------------- */
%  orthogonalize in order given */
% ----------------------------- */

   r = chol(sigma)';
   rinv = inv(r);

   if bigres;
       ARorth = rinv*AR;
       Gorth =  rinv*G;
       muorth = rinv*mu;
   end;

%% find impulse response function         
% this uses the nonorthogonalized        
% representation, finds the e innovation 
% corresponding to a unit v innov.       

   % ----------------------------------------*/
   % simulate ols system in companion form   */
   % starting from sample means.             */
   % simulation ignores means.               */
   % [ Dy(t) ] =  [ b' ] [ y(t-1) - c(t-1) ]   +  [ e1 ]
   % [ Dc(t) ]           [ Dx(t-1)         ]      [ e2 ]
   %                     [ Dx(t-lag)       ]
   %                     [ Dy(t-1)         ]
   %                     [ Dy(t-lag)       ]
   %
   %   xsim              compan
   %
   %  start from y = c = 0 and all diffs = 0 */
   % ----------------------------------------*/

   shock = 1;                                 % which series gets shock? */
   while shock <= nseries;
        dosim;
        if shock == 1;
            Dxir = Dxsim;
            xir = xsim;
        else
            Dxir = [Dxir Dxsim];
            xir = [xir xsim];
        end;
        shock = shock+1;
   end;

   vardec = recserar(Dxir.^2,Dxir(1,:).^2,ones(1,4));
                         % this is decomposition of forecast error variance */

%% find beveridge-nelson decomposition  
% again, you can simply ignore means   

% form companion matrix: updates Rhv(t)' = A*Rhv(t-1)'+ error */
  if bigres;
    statcom = 0;
    if cols(x) == 2;  % not yet implemented for more right hand vars */

        arows = rows(b)-1;
        bnoconstant = b(2:rows(b),:);
        Amat = zeros(arows,arows);
        Amat(1,:) =  a'*(bnoconstant') + [1 zeros(1,arows-1)];     % ONE  ci vect */
        if lags >= 2;
            j = 1;
            while j <= nseries;
            Amat((2+(j-1)*lags),:) =  bnoconstant(:,j)';           % Dxj(t)       */
                ll = (3+(j-1)*lags) ;
                lr = (3+(j-1)*lags+lags-2);
                rl = (2+(j-1)*lags);
                rr = (2+(j-1)*lags+lags-2);
                Amat(ll:lr,rl:rr) = eye(lags-1);
                j = j+1;
            end;
            cofst = [zeros(cols(a)+lags,1);1;zeros(lags-1,1)];
        end;

        if lags == 1;
            Amat(2:arows,:) = bnocon';
            cofst = zeros(cols(a)+1,1)|1;
        end;

        cofst = cofst'*Amat*inv(eye(arows)-Amat);
        cofst = [0 cofst];
                                                     % Rhv(t) is t-1 info ! */
        statcom = -(Rhv(2:rows(Rhv),:)-repmat(mean(Rhv),size(Rhv(2:rows(Rhv),1)),1))*cofst';
        lastRh = [1 (x(rows(x),:)*a)];
        if lags > 0;                                % then lags of Dx */
            i = 1;
            while i <= cols(xl);
                j = 1;
                while j <= lags;
                    lastRh = [lastRh Dx(T+1-j,i)];
                    j = j+1;
                end;
                i = i+1;
            end;
        end;

        statcom = [statcom ; ...
                  (-(lastRh - mean(Rhv))*cofst')];

    end;
  end;

%% augmented Dickey-Fuller tests on 
% ci vector                        

    if bigres;
        
        fprintf('(Augmented Dickey fuller regressions for d-p)');
        fprintf('D(d-p)(t) = const. + rho(d-p)(t-1) + b1 D (d-p) (t-1) + ... ');
%         format /RD 9,3;

        ci = xl*a;
        dci = ci(2:T)-ci(1:T-1);      % ci is lagged rel. to dci */

        j = 0;
        while j <= 5; % max amount of augmenting */
            lhv = dci(1+j:T-1);
            RhvDF = [ ones(T-1-j,1) ci(1+j:T-1)];
            if j > 0;
                i = 1;
                while i <= j;
                    RhvDF = [RhvDF dci(1+j-i:T-i-1)];
                i = i+1;
                end;
            end;
            bci = RhvDF\lhv;
            eci = lhv-RhvDF*bci;
            sigci = eci'*eci/(T-j-1);
            sebci = (diag(inv(RhvDF'*RhvDF)).*diag(sigci)').^(1/2);
            tci = bci./sebci;

            if j == 0;
                dfcon = 1;
            end;
            if j > 0;
                dfcon = 1-sumc(bci(3:rows(bci)));
            end;
            rhotest = bci(2)/dfcon*(T-1);

            fprintf('\n \n');
            fprintf('coeff.  %8.2f %8.2f \n', bci');
            fprintf('se      %8.2f %8.2f \n', sebci');
            fprintf('tstat   %8.2f %8.2f \n', tci');
            fprintf('rhostat %8.2f \n', rhotest);
            j = j+1;
        end;
    end;

%%    print table of results      
%  some have been commented out  


 if prnt;
   fprintf('VAR estimates\n');
   
    % ----------------------------------- */
    % create header to identify variables */
    % ----------------------------------- */

    blank = {'(t-1)'; '(t-2)'; '(t-3)'; '(t-4)'; '(t-5)'; '(t-6)'; '(t-7)'; '(t-8)'};
    header ={'    '; 'const'};
%     j = 1;
%     while j <= cols(a);
%         header{size(header,1)+1} = strcat(cellstr(names(j)),'t-1');     % add names of ci vectors */
%         j = j+1;
%     end;

%     i = cols(a) + 1;
%     while i <= rows(names);
%         j = 1;
%         while j <= lags;
%             header = header|(names[i] + blank[j]);
%             header{size(header,1)+1} = strcat(cellstr(names(i)),blank{j});
%             j = j+1;
%         end;
%         i = i+1;
%     end;
    header ={'const';'d-p_t-1';'d_t-1';'d_t-2';'p_t-1';'p_t-2'};
    
    % -------------------------- */
    % create table of ols coeffs */
    % -------------------------- */

%     table = header'|
%             ((names[cols(a)+1:rows(names)] $+ '(t)') ~ b');
%     mask = zeros(1,cols(table))|(zeros(rows(b'),1)~ones(rows(b'),cols(b')));

    fprintf('OLS coefficients\n');
    for ii=1:cols(Rhv)
        fprintf('%s %8.2f %8.2f\n',header{ii},b(ii,:))
    end
%     printfm(table, mask, fmt);


    fprintf('Standard Errors\n');
%     table = header'|
%             ((names[cols(a)+1:rows(names)] $+ "(t)") ~ seb');
%     printfm(table, mask, fmt);
    for ii=1:cols(Rhv)
        fprintf('%s %8.2f %8.2f\n',header{ii},seb(ii,:))
    end


    fprintf('t Ratios\n');
%     table = header'|
%             ((names[cols(a)+1:rows(names)] $+ '(t)') ~ (b'./seb'));
%     printfm(table, mask, fmt);
    for ii=1:cols(Rhv)
        fprintf('%s %8.2f %8.2f\n',header{ii},b(ii,:)./seb(ii,:))
    end


    fprintf('R2, pvalue\n');
    fprintf('%8.2f %8.2f\n',R2, pvf);

    fprintf('Variance/covariance, correlation matrix of errors\n');

    ss = diag(sigma).^(1/2);
    correl = sigma ./ (ss*ss');
    fprintf('sigma \n'); 
    disp(sigma);
    fprintf('corr \n'); 
    disp(correl);

    % ------------------------------------ */
    % AR and orthogonalized representation */
    % ------------------------------------ */
    if bigres;
        fprintf('A(L)Dx(t) = mu + G(a''x(t-1)) + e representation. \n');
        fprintf('A(L): A0 ~ A1 ~ A2 ..."');
%         format /RD 9,3;
        AR
        fprintf('G'); disp(G)
        fprintf('mu');disp(mu)

        fprintf('A(L)Dx(t) = mu + G(a''x(t-1)) + v orthogonalized representation. \n');
        fprintf('This is the same as ols with current dividend in y \n');
        fprintf('A(L): A0 ~ A1 ~ A2 ..."');
%         format /RD 9,3;
        ARorth
        fprintf('G'); disp(Gorth)
        fprintf('mu');disp(muorth)
    end;

    % ------------------------------------------------ */
    % print variance-covariance and correlation matrix */
    % ------------------------------------------------ */

    
    fprintf('Variance covariance matrix\n');
    disp(sigma);

    fprintf('Correlation matrix\n');
    d = diag(sigma);
    disp(sigma./((d*d')^(1/2)));

    fprintf('choleski decomp of sigma\n');
    disp(r);
 

 %% print impulse-response functions  
  
% 
%     "Impulse response function";
%     "Differences";
%     "   x1 to x1 shock, x2 to x1 shock.. x1 to x2 shock x2 to x2 shock..";
%     "Differences: Dx1 Dx2.. to shock in x1,x2,...";
%     format /RD 10,4; Dxir;
%     "";
%     "Levels";
%     "   x1 to x1 shock, x2 to x1 shock.. x1 to x2 shock x2 to x2 shock.. ";
%     format /RD 10,4; xir;
% 
    
    denom = kron([1 1], std(Dx));
    fprintf('Decomposition of variance\n');
    fprintf('order:x1 to x1 shock, x2 to x1 shock ... x1 to x2 shock x2 to x2 shock..\n');
    fprintf('1: one step ahead forecast error variance %8.3f %8.3f %8.3f %8.3f\n',vardec(1,:));
    fprintf('variance ratio to series variance x 100:  %8.3f %8.3f %8.3f %8.3f \n', 100*vardec(1,:)./(denom.^2));

    fprintf('\n \n');
    fprintf('2: horiz step ahead forecast error = series variance: %8.3f %8.3f %8.3f %8.3f\n', vardec(rows(vardec),:));
    fprintf('variance ratio to series variance x 100:              %8.3f %8.3f %8.3f %8.3f \n', 100*vardec(rows(vardec),:)./(denom.^2));

    fprintf('series variance:            %8.3f %8.3f %8.3f %8.3f\n',denom.^2);
    fprintf('series standard deviation:  %8.3f %8.3f %8.3f %8.3f\n',denom);
    end;   % ends if on print section */
